import boto3
import tenacity
import taskqueue
import time

from taskqueue import RegisteredTask as TQRegisteredTask

import mazepa

retry = tenacity.retry(
  reraise=True,
  stop=tenacity.stop_after_attempt(4),
  wait=tenacity.wait_full_jitter(0.5, 60.0),
  )


class BasicQueue:
    def submit_task_batch(self):
        raise NotImplementedError

    def get_completed(self):
        raise NotImplementedError


# TQ wrapper. Theretically don't have to use TQ library, but it's nice
class MazepaTaskTQ(TQRegisteredTask):
    def __init__(self, task_spec):
        if not isinstance(task_spec, str):
            task_spec = mazepa.serialize(task_spec)
        super().__init__(task_spec=task_spec)

    def execute(self):
        task = mazepa.deserialize(self.task_spec)
        print (f"Starting execution of {type(task)}")
        ts = time.time()
        task.execute()
        te = time.time()
        print (f"Done! Execution time: {te - ts} seconds.")

class Queue(BasicQueue):
    def __init__(self, queue_name=None, threads=1, queue_region='us-east-1'):
        self.threads = threads
        self.queue_name = queue_name
        if queue_name is None:
            self.local_execution = True
            #self.queue = taskqueue.LocalTaskQueue(parallel=1)
        else:
            self.local_execution = False
            self.queue = taskqueue.GreenTaskQueue(queue_name)
            self.queue_boto = boto3.client('sqs',
                                            region_name=queue_region)
            self.queue_url = self.queue_boto.get_queue_url(QueueName=self.queue_name)["QueueUrl"]

    def submit_tq_tasks(self, tq_tasks):
        if self.threads > 1:
            #TODO
            raise NotImplementedError
        self.queue.insert_all(tq_tasks)

    def submit_mazepa_tasks(self, mazepa_tasks):
        if self.threads > 1:
            #TODO
            raise NotImplementedError
        if self.local_execution:
            print ("Starting local task execution...")
            for t in mazepa_tasks:
                t.execute()
        else:
            tq_tasks = [MazepaTaskTQ(t) for t in mazepa_tasks]
            self.submit_tq_tasks(tq_tasks)

    @retry
    def remote_queue_is_empty(self):
        """Is our remote queue empty?
        """
        #TODO: cleanup
        attribute_names = ['ApproximateNumberOfMessages', 'ApproximateNumberOfMessagesNotVisible']
        responses = []
        for i in range(3):
            response = self.queue_boto.get_queue_attributes(
                    QueueUrl=self.queue_url,
                    AttributeNames=attribute_names)
            for a in attribute_names:
                responses.append(int(response['Attributes'][a]))
                print('{}     '.format(responses[-2:]), end="\r", flush=True)
            if i < 2:
              time.sleep(1)

        return all(n == 0 for n in responses)

    def is_local_queue(self):
        return self.local_execution

    def is_remote_queue(self):
        return not self.is_local_queue()

    def get_completed(self):
        if self.is_local_queue() or self.remote_queue_is_empty():
            return mazepa.job.AllJobsIndicator()
        else:
            return None

    def poll_tasks(self, lease_seconds):
        assert self.is_remote_queue
        self.queue.poll(lease_seconds=lease_seconds)







class Task:
    def __init__(self, *kargs, **kwargs):
        self.tags = {}

    def __call__(self, *kargs, **kwargs):
        raise NotImplementedError


class Job:
    def __init__(self, *kargs, **kwargs):
        pass

    def get_tasks(self):
        raise NotImplementedError

    def get_next_task_batch(self):
        '''
        returns: list of tasks that can be completed. when done, returns an empty list
        '''
        result = []
        while True:
            #squeeze the job until it's either done or returns a barrier
            try:
                next_step = self.get_tasks()
            except StopIteration:
                #means that this job has no more tasks, it's done
                break
            else:
                if next_step is Barrier:
                    if len(result) == 0:
                        raise Exception(f"Job '{type(self)}' issued two Bariers in a row")
                    break
                elif isinstance(next_step, list):
                    result.extend(next_step)

        return result

    def old_get_next_task_batch(self):
            '''
            Main execution loop.
            Returns type of next(self.iterator) must be either a lits (of tasks),
            execution signal, or another Mazepa Job. When the job is done,
            next(self.iterator)) will raise a StopIteration exception.

            Results are accumulated until one of two conditions is met:
                1)  next(self.iterator)) raises a StopIteration exception, meaning
                   that this job is complete, and the no_more_tasks flag is set
                   to indicate completion.
                2) next(self.iterator) returns an execution signal of type Barrier,
                   which means that proceeding tasks need to wait for current
                   tasks to complete first.
            '''
            while True:
                try:
                    if isinstance(self.next_step, MazepaExecutionSignal):
                        if isinstance(self.next_step, Barrier):
                            self.next_step = next(self.iterator)
                            return result
                        else:
                            raise Exception("Unknown execution signal \
                                    of type {}".format(type(self.next_step)))

                    elif isinstance(self.next_step, Job):
                        subjob = self.next_step
                        subjob_task_batch = subjob.get_next_task_batch()
                        result += subjob_task_batch

                        if subjob.has_more_tasks():
                            # if the subjob is not done, it means theres
                            # a dependency inside it, ie it received a Barrier.
                            # have to wait until current tasks complete
                            # the next iteration of the while loop will come back
                            # to call get_next_task_batch on this subjob
                            return result
                        else:
                            self.next_step = next(self.iterator)

                    else:
                        task_list = self.next_step
                        result += task_list
                        self.next_step = next(self.iterator)

                except StopIteration:
                   self.no_more_tasks = True
                   return result

class MazepaExecutionSignal:
    def __init__(self):
        pass

class Barrier(MazepaExecutionSignal):
    def __init__(self):
        pass

class JobReadyIndicator:
    def __init__(self):
        pass

class JobReadyList(JobReadyIndicator):
    def __init__(self, job_list):
        super().__init__()
        self.job_list = job_list

class AllJobsIndicator(JobReadyIndicator):
    def __init__(self):
        super().__init__()




import pickle
import codecs

def serialize(obj):
    return codecs.encode(pickle.dumps(obj, protocol=4), "base64").decode()

def deserialize(s):
    return pickle.loads(codecs.decode(s.encode(), "base64"))
